//===========================================================================
//  @(#) $Name: arts++-1-1-a8 $
//  @(#) $Id: ArtsIpPathData.cc,v 1.17 2000/10/02 17:05:25 dwm Exp $
//===========================================================================
//  Copyright Notice
//
//  By accessing this software, arts++, you are duly informed
//  of and agree to be bound by the conditions described below in this
//  notice:
//
//  This software product, arts++, is developed by Daniel W. McRobb, and
//  copyrighted(C) 1998 by the University of California, San Diego
//  (UCSD), with all rights reserved.  UCSD administers the CAIDA grant,
//  NCR-9711092, under which part of this code was developed.
//
//  There is no charge for arts++ software.  You can redistribute it
//  and/or modify it under the terms of the GNU Lesser General Public
//  License, Version 2.1, February 1999, which is incorporated by
//  reference herein.
//
//  arts++ is distributed WITHOUT ANY WARRANTY, IMPLIED OR EXPRESS, OF
//  MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE or that the use
//  of it will not infringe on any third party's intellectual
//  property rights.
//
//  You should have received a copy of the GNU Lesser General Public
//  License along with arts++.  Copies can also be obtained from:
//
//    http://www.gnu.org/copyleft/lesser.html
//
//  or by writing to:
//
//  Free Software Foundation, Inc.
//  59 Temple Place, Suite 330
//  Boston, MA 02111-1307
//  USA
//
//  Or contact:
//
//    info@caida.org
//===========================================================================

extern "C" {
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
}

#include <string>
#include <algorithm>

#include "Arts.hh"
#include "ArtsIpPathData.hh"

static const string rcsid = "@(#) $Name: arts++-1-1-a8 $ $Id: ArtsIpPathData.cc,v 1.17 2000/10/02 17:05:25 dwm Exp $";

//------------------------------------------------------------------------
//     ArtsIpPathData::ArtsIpPathData(ipv4addr_t src, ipv4addr_t dst)
//------------------------------------------------------------------------
ArtsIpPathData::ArtsIpPathData(ipv4addr_t src, ipv4addr_t dst)
{
  this->Clear();
  
  this->_src = src;
  this->_dst = dst;

  #ifndef NDEBUG
    ++this->_numObjects;
  #endif
}

//-------------------------------------------------------------------------
//  ArtsIpPathData::ArtsIpPathData(const ArtsIpPathData & artsIpPathData)  
//.........................................................................
//                                                                         
//-------------------------------------------------------------------------
ArtsIpPathData::ArtsIpPathData(const ArtsIpPathData & artsIpPathData)
{
  this->_src = artsIpPathData._src;
  this->_dst = artsIpPathData._dst;
  this->_rtt = artsIpPathData._rtt;
  this->_hopDistance = artsIpPathData._hopDistance;
  this->_isComplete = artsIpPathData._isComplete;
  this->_numHops = artsIpPathData._numHops;
  this->_incompleteReason = artsIpPathData._incompleteReason;
  this->_incompleteFlags = artsIpPathData._incompleteFlags;
  this->_path = artsIpPathData._path;

  #ifndef NDEBUG
    ++this->_numObjects;
  #endif
}

//------------------------------------------------------------------------
//
//------------------------------------------------------------------------
ArtsIpPathData::ArtsIpPathData()
{
  this->Clear();

  #ifndef NDEBUG
    ++this->_numObjects;
  #endif
}

//----------------------------------------------------------------------------
//     uint8_t ArtsIpPathData::LoopLength() const 
//............................................................................
//  
//----------------------------------------------------------------------------
uint8_t ArtsIpPathData::LoopLength() const
{
  uint8_t  rc = 0;
  if (this->_incompleteReason == ArtsIpPathData::k_loopDetected)
    rc = this->_incompleteFlags;
  return(rc);
}

//----------------------------------------------------------------------------
//     uint8_t ArtsIpPathData::LoopLength(uint8_t loopLength) 
//............................................................................
//  
//----------------------------------------------------------------------------
uint8_t ArtsIpPathData::LoopLength(uint8_t loopLength)
{
  this->_incompleteReason = ArtsIpPathData::k_loopDetected;
  this->_incompleteFlags = loopLength;
  return(this->_incompleteFlags);
}

//----------------------------------------------------------------------------
//     uint8_t ArtsIpPathData::TtlLimit() const 
//............................................................................
//  
//----------------------------------------------------------------------------
uint8_t ArtsIpPathData::TtlLimit() const
{
  uint8_t  rc = 0;
  if (this->_incompleteReason == ArtsIpPathData::k_ttlLimit)
    rc = this->_incompleteFlags;
  return(rc);
}

//----------------------------------------------------------------------------
//     uint8_t ArtsIpPathData::TtlLimit(uint8_t ttlLimit) 
//............................................................................
//  
//----------------------------------------------------------------------------
uint8_t ArtsIpPathData::TtlLimit(uint8_t ttlLimit)
{
  this->_incompleteReason = ArtsIpPathData::k_ttlLimit;
  this->_incompleteFlags = ttlLimit;
  return(this->_incompleteFlags);
}

//----------------------------------------------------------------------------
//     void ArtsIpPathData::Clear() 
//............................................................................
//  
//----------------------------------------------------------------------------
void ArtsIpPathData::Clear()
{
  this->_src = 0;
  this->_dst = 0;
  this->_hopDistance = 0;
  this->_isComplete = 0;
  this->_numHops = 0;
  this->_incompleteReason = 0;
  this->_incompleteFlags = 0;
  
  return;
}

//------------------------------------------------------------------------
//                      ArtsIpPathData::~ArtsIpPathData()
//------------------------------------------------------------------------
ArtsIpPathData::~ArtsIpPathData()
{
  #ifndef NDEBUG
    --this->_numObjects;
  #endif
}

//----------------------------------------------------------------------------
//        uint32_t ArtsIpPathData::Length(uint8_t version = 0) const 
//............................................................................
//  
//----------------------------------------------------------------------------
uint32_t ArtsIpPathData::Length(uint8_t version = 0) const
{
  uint32_t len =  sizeof(this->_src) + sizeof(this->_dst) +
    sizeof(this->_rtt) + sizeof(this->_hopDistance) + sizeof(this->_numHops);

  if (version == 1) {
    if (! this->_isComplete) {
      len += sizeof(this->_incompleteReason) + sizeof(this->_incompleteFlags);
    }
  }
  
  vector<ArtsIpPathEntry>::const_iterator  pathEntry;
  for (pathEntry = this->Path().begin(); pathEntry != this->Path().end();
       pathEntry++) {
    len += pathEntry->Length(version);
  }
  return(len);
}

//------------------------------------------------------------------------
//   ostream& ArtsIpPathData::write(ostream& os, uint8_t version = 0)
//------------------------------------------------------------------------
ostream& ArtsIpPathData::write(ostream& os, uint8_t version = 0)
{
  uint32_t  timeDatum;
  uint8_t   completeAndNumHops;

  os.write(&this->_src,sizeof(this->_src));
  os.write(&this->_dst,sizeof(this->_dst));

  timeDatum = htonl(this->_rtt.tv_sec);
  os.write(&timeDatum,sizeof(timeDatum));
  timeDatum = htonl(this->_rtt.tv_usec);
  os.write(&timeDatum,sizeof(timeDatum));

  os.write(&this->_hopDistance,sizeof(this->_hopDistance));
  
  completeAndNumHops = (this->_isComplete << 7) | this->_numHops;
  os.write(&completeAndNumHops,sizeof(completeAndNumHops));

  if (version == 1) {
    if (! this->_isComplete) {
      os.write(&this->_incompleteReason,sizeof(this->_incompleteReason));
      os.write(&this->_incompleteFlags,sizeof(this->_incompleteFlags));
    }
  }

  //  sort by hop number
  sort(this->_path.begin(),this->_path.end(),less<ArtsIpPathEntry>());

  //  write each hop
  vector<ArtsIpPathEntry>::const_iterator  pathEntry;
  for (pathEntry = this->_path.begin(); pathEntry != this->_path.end();
       pathEntry++) {
    (*pathEntry).write(os,version);
  }
  return(os);
}

//-------------------------------------------------------------------------
//          int ArtsIpPathData::write(int fd, uint8_t version = 0)
//.........................................................................
//                                                                         
//-------------------------------------------------------------------------
int ArtsIpPathData::write(int fd, uint8_t version = 0)
{
  uint32_t  timeDatum;
  uint8_t   completeAndNumHops;
  int       rc;
  int       bytesWritten = 0;
  
  rc = g_ArtsLibInternal_Primitive.FdWrite(fd,&this->_src,sizeof(this->_src));
  if (rc != sizeof(this->_src)) {
    return(-1);
  }
  bytesWritten += rc;
  
  rc = g_ArtsLibInternal_Primitive.FdWrite(fd,&this->_dst,sizeof(this->_dst));
  if (rc != sizeof(this->_dst)) {
    return(-1);
  }
  bytesWritten += rc;

  timeDatum = htonl(this->_rtt.tv_sec);
  rc = g_ArtsLibInternal_Primitive.FdWrite(fd,&timeDatum,sizeof(timeDatum));
  if (rc != sizeof(timeDatum)) {
    return(-1);
  }
  bytesWritten += rc;
  
  timeDatum = htonl(this->_rtt.tv_usec);
  rc = g_ArtsLibInternal_Primitive.FdWrite(fd,&timeDatum,sizeof(timeDatum));
  if (rc != sizeof(timeDatum)) {
    return(-1);
  }
  bytesWritten += rc;

  rc = g_ArtsLibInternal_Primitive.FdWrite(fd,&this->_hopDistance,
                                           sizeof(this->_hopDistance));
  if (rc != sizeof(this->_hopDistance)) {
    return(-1);
  }
  bytesWritten += rc;
  
  completeAndNumHops = (this->_isComplete << 7) | this->_numHops;
  rc = g_ArtsLibInternal_Primitive.FdWrite(fd,&completeAndNumHops,
                                           sizeof(completeAndNumHops));
  if (rc != sizeof(completeAndNumHops)) {
    return(-1);
  }
  bytesWritten += rc;

  if (version == 1) {
    if (! this->_isComplete) {
      rc =
        g_ArtsLibInternal_Primitive.FdWrite(fd,&this->_incompleteReason,
                                            sizeof(this->_incompleteReason));
      if (rc != sizeof(this->_incompleteReason))
        return(-1);
      bytesWritten += rc;

      rc = g_ArtsLibInternal_Primitive.FdWrite(fd,&this->_incompleteFlags,
                                               sizeof(this->_incompleteFlags));
      if (rc != sizeof(this->_incompleteFlags))
        return(-1);
      bytesWritten += rc;
    }
  }
  
  //  sort by hop number
  sort(this->_path.begin(),this->_path.end(),less<ArtsIpPathEntry>());

  //  write each hop
  vector<ArtsIpPathEntry>::const_iterator  pathEntry;
  for (pathEntry = this->_path.begin(); pathEntry != this->_path.end();
       pathEntry++) {
    if ((rc = (*pathEntry).write(fd,version)) < 0) {
      return(-1);
    }
    bytesWritten += rc;
  }
  return(bytesWritten);
}

//------------------------------------------------------------------------
//    istream& ArtsIpPathData::read(istream& is, uint8_t version = 0)
//------------------------------------------------------------------------
istream& ArtsIpPathData::read(istream& is, uint8_t version = 0)
{
  uint32_t  timeDatum;
  uint8_t   completeAndNumHops;
  size_t    hopNum;
  
  is.read(&this->_src,sizeof(this->_src));
  is.read(&this->_dst,sizeof(this->_dst));

  is.read(&timeDatum,sizeof(timeDatum));
  this->_rtt.tv_sec = ntohl(timeDatum);
  is.read(&timeDatum,sizeof(timeDatum));
  this->_rtt.tv_usec = ntohl(timeDatum);

  is.read(&this->_hopDistance,sizeof(this->_hopDistance));
  
  is.read(&completeAndNumHops,sizeof(completeAndNumHops));
  this->_isComplete = completeAndNumHops >> 7;
  this->_numHops = completeAndNumHops & 0x7f;

  if (version == 1) {
    if (! this->_isComplete) {
      is.read(&this->_incompleteReason,sizeof(this->_incompleteReason));
      is.read(&this->_incompleteFlags,sizeof(this->_incompleteFlags));
    }
  }
  
  if (this->_path.size() > 0) {
    this->_path.erase(this->_path.begin(),this->_path.end());
  }
  
  this->_path.reserve(this->_numHops);

  //  read in contents for each hop

  ArtsIpPathEntry  pathEntry;
  
  for (hopNum = 0; hopNum < this->_numHops; hopNum++) {
    pathEntry.read(is,version);
    this->_path.push_back(pathEntry);
  }
  
  return(is);
}

//-------------------------------------------------------------------------
//           int ArtsIpPathData::read(int fd, uint8_t version = 0)
//.........................................................................
//                                                                         
//-------------------------------------------------------------------------
int ArtsIpPathData::read(int fd, uint8_t version = 0)
{
  uint32_t  timeDatum;
  uint8_t   completeAndNumHops;
  size_t    hopNum;
  int       rc;
  int       bytesRead = 0;
  
  rc = g_ArtsLibInternal_Primitive.FdRead(fd,&this->_src,sizeof(this->_src));
  if (rc <= 0)
    return(rc);
  bytesRead += rc;
  rc = g_ArtsLibInternal_Primitive.FdRead(fd,&this->_dst,sizeof(this->_dst));
  if (rc <= 0)
    return(rc);
  bytesRead += rc;
  
  rc = g_ArtsLibInternal_Primitive.FdRead(fd,&timeDatum,sizeof(timeDatum));
  if (rc <= 0)
    return(rc);
  bytesRead += rc;
  this->_rtt.tv_sec = ntohl(timeDatum);
  rc = g_ArtsLibInternal_Primitive.FdRead(fd,&timeDatum,sizeof(timeDatum));
  if (rc <= 0)
    return(rc);
  bytesRead += rc;
  this->_rtt.tv_usec = ntohl(timeDatum);

  rc = g_ArtsLibInternal_Primitive.FdRead(fd,&this->_hopDistance,
                                          sizeof(this->_hopDistance));
  if (rc <= 0)
    return(rc);
  bytesRead += rc;
  
  rc = g_ArtsLibInternal_Primitive.FdRead(fd,&completeAndNumHops,
                                          sizeof(completeAndNumHops));
  if (rc <= 0)
    return(rc);
  bytesRead += rc;
  this->_isComplete = completeAndNumHops >> 7;
  this->_numHops = completeAndNumHops & 0x7f;

  if (version == 1) {
    if (! this->_isComplete) {
      rc = g_ArtsLibInternal_Primitive.FdRead(fd,&this->_incompleteReason,
                                              sizeof(this->_incompleteReason));
      if (rc <= 0)
        return(rc);
      bytesRead += rc;
      
      rc = g_ArtsLibInternal_Primitive.FdRead(fd,&this->_incompleteFlags,
                                              sizeof(this->_incompleteFlags));
      if (rc <= 0)
        return(rc);
      bytesRead += rc;
    }
  }
  
  if (this->_path.size() > 0) {
    this->_path.erase(this->_path.begin(),this->_path.end());
  }
  
  this->_path.reserve(this->_numHops);

  //  read in contents for each hop

  ArtsIpPathEntry  pathEntry;
  
  for (hopNum = 0; hopNum < this->_numHops; hopNum++) {
    rc = pathEntry.read(fd,version);
    if (rc <= 0)
      return(rc);
    bytesRead += rc;
    this->_path.push_back(pathEntry);
  }
  
  return(bytesRead);
}

//-------------------------------------------------------------------------
//     bool                                  
//     ArtsIpPathData::
//     Distinguishable(const vector<ArtsIpPathEntry> & ipPath) const 
//.........................................................................
//  
//-------------------------------------------------------------------------
bool
ArtsIpPathData::Distinguishable(const vector<ArtsIpPathEntry> & ipPath) const
{
  vector<ArtsIpPathEntry>::const_iterator  thisHopIter;
  vector<ArtsIpPathEntry>::const_iterator  inHopIter;

  for (thisHopIter = (*this)._path.begin();
       thisHopIter != (*this)._path.end(); thisHopIter++) {
    for (inHopIter = ipPath.begin();
         inHopIter != ipPath.end(); inHopIter++) {
      if ((*thisHopIter).HopNum() == (*inHopIter).HopNum()) {
        if ((*thisHopIter).IpAddr() != (*inHopIter).IpAddr()) {
          return(true);
        }
      }
    }
  }
  return(false);
}

//-------------------------------------------------------------------------
// bool                                  
// ArtsIpPathData::Distinguishable(const ArtsIpPathData & ipPathData) const 
//.........................................................................
//  
//-------------------------------------------------------------------------
bool
ArtsIpPathData::Distinguishable(const ArtsIpPathData & ipPathData) const
{
  if (this->_src != ipPathData.Src() || this->_dst != ipPathData.Dst()) {
    //  different source or destination addresses
    return(true);
  }
  
  if (this->_isComplete && ipPathData.IsComplete()) {
    //  both paths claim to be complete; compare the hop distance
    if (this->_hopDistance != ipPathData.HopDistance()) {
      return(true);
    }
  }
  
  if (this->Distinguishable(ipPathData.Path())) {
    //  the hop vectors are distinguishable
    return(true);
  }
  return(false);
}

//-------------------------------------------------------------------------
//        vector<ipv4addr_t> * ArtsIpPathData::HopAddresses() const        
//.........................................................................
//  
//-------------------------------------------------------------------------
vector<ipv4addr_t> * ArtsIpPathData::HopAddresses() const
{
  vector<ipv4addr_t>  *newAddrVect = new vector<ipv4addr_t>;

  vector<ArtsIpPathEntry>::const_iterator  thisHopIter;
  for (thisHopIter = (*this)._path.begin();
       thisHopIter != (*this)._path.end(); thisHopIter++) {
    (*newAddrVect).push_back((*thisHopIter).IpAddr());
  }
  return(newAddrVect);
}

//-------------------------------------------------------------------------
//     bool                                  
//     ArtsIpPathData::
//     CommonHopAddresses(const vector<ArtsIpPathEntry> & ipPath, 
//                        vector<ipv4addr_t> & commonHopAddrs) const 
//.........................................................................
//  
//-------------------------------------------------------------------------
bool
ArtsIpPathData::CommonHopAddresses(const vector<ArtsIpPathEntry> & ipPath,
                                   vector<ipv4addr_t> & commonHopAddrs) const
{
  vector<ArtsIpPathEntry>::const_iterator  thisHopIter;
  vector<ArtsIpPathEntry>::const_iterator  inHopIter;
  bool                                     rc = false;
  
  for (thisHopIter = (*this)._path.begin();
       thisHopIter != (*this)._path.end(); thisHopIter++) {
    for (inHopIter = ipPath.begin(); inHopIter != ipPath.end(); inHopIter++) {
      if ((*thisHopIter).IpAddr() == (*inHopIter).IpAddr()) {
        rc = true;
        if (find(commonHopAddrs.begin(),commonHopAddrs.end(),
                 (*thisHopIter).IpAddr()) == commonHopAddrs.end()) {
          commonHopAddrs.push_back((*thisHopIter).IpAddr());
        }
      }
    }
  }
  return(rc);
}

//-------------------------------------------------------------------------
//    bool                                  
//    ArtsIpPathData::                      
//    CommonHopAddresses(const vector<ipv4addr_t> & hopAddresses, 
//                       vector<ipv4addr_t> & commonHopAddresses) const 
//.........................................................................
//  
//-------------------------------------------------------------------------
bool
ArtsIpPathData::
CommonHopAddresses(const vector<ipv4addr_t> & hopAddresses,
                   vector<ipv4addr_t> & commonHopAddresses) const
{
  vector<ArtsIpPathEntry>::const_iterator  thisHopIter;
  bool                                     rc;

  if (commonHopAddresses.size() > 0) {
    commonHopAddresses.erase(commonHopAddresses.begin(),
                             commonHopAddresses.end());
  }
  
  for (thisHopIter = (*this)._path.begin();
       thisHopIter != (*this)._path.end(); thisHopIter++) {
    if (find(hopAddresses.begin(),hopAddresses.end(),(*thisHopIter).IpAddr())
        != hopAddresses.end()) {
      rc = true;
      if (find(commonHopAddresses.begin(),commonHopAddresses.end(),
               (*thisHopIter).IpAddr()) == commonHopAddresses.end()) {
        commonHopAddresses.push_back((*thisHopIter).IpAddr());
      }
    }
  }
  return(rc);
}


//-------------------------------------------------------------------------
// void
// ArtsIpPathData::MergeWithPath(const vector<ArtsIpPathEntry> & ipPath,  
//                               vector<ArtsIpPathEntry> & mergedPath) const 
//.........................................................................
//  
//-------------------------------------------------------------------------
void
ArtsIpPathData::MergeWithPath(const vector<ArtsIpPathEntry> & ipPath,
                              vector<ArtsIpPathEntry> & mergedPath) const
{
  vector<ArtsIpPathEntry>::iterator       unionIterBegin, unionIterEnd;
  ArtsIpPathEntryLessByHopNumber          hopsLess;
  
  mergedPath.reserve((*this)._path.size() + ipPath.size());
  unionIterBegin = mergedPath.begin();
  unionIterEnd = set_union((*this)._path.begin(),(*this)._path.end(),
                           ipPath.begin(),ipPath.end(),
                           unionIterBegin,hopsLess);
  mergedPath.insert(mergedPath.begin(),unionIterBegin,unionIterEnd);
  return;
}


//-------------------------------------------------------------------------
//   ArtsIpPathData &
//   ArtsIpPathData::operator = (const ArtsIpPathData & artsIpPathData)
//.........................................................................
//  
//-------------------------------------------------------------------------
ArtsIpPathData &
ArtsIpPathData::operator = (const ArtsIpPathData & artsIpPathData)
{
  this->_src = artsIpPathData._src;
  this->_dst = artsIpPathData._dst;
  this->_rtt = artsIpPathData._rtt;
  this->_hopDistance = artsIpPathData._hopDistance;
  this->_isComplete = artsIpPathData._isComplete;
  this->_numHops = artsIpPathData._numHops;
  this->_incompleteReason = artsIpPathData._incompleteReason;
  this->_incompleteFlags = artsIpPathData._incompleteFlags;
  this->_path = artsIpPathData._path;
  return(*this);
}

//------------------------------------------------------------------------
//
//------------------------------------------------------------------------
ostream& operator << (ostream& os, const ArtsIpPathData & artsIpPathData)
{
  struct in_addr  inAddr;
  
  os << "IPPATH OBJECT DATA" << endl;
  inAddr.s_addr = artsIpPathData.Src();
  os << "\tSrc: " << setiosflags(ios::left) << setw(16) << inet_ntoa(inAddr)
     << setiosflags(ios::showbase) << " (" << setw(8) << hex
     << artsIpPathData.Src() << ")" << endl;
  inAddr.s_addr = artsIpPathData.Dst();
  os << "\tDst: " << setiosflags(ios::left) << setw(16) << inet_ntoa(inAddr)
     << " (" << setiosflags(ios::showbase) << setw(8) << hex
     << artsIpPathData.Dst() << ")" << endl;
  os << "\tRtt: " << (float)((artsIpPathData.Rtt().tv_sec * 1000000.0 +
                              artsIpPathData.Rtt().tv_usec)/1000.0)
     << " ms" << endl;
  os << "\tHopDistance: " << dec << (int)artsIpPathData.HopDistance()
     << " (" << hex << (int)artsIpPathData.HopDistance() << ")" << endl;
  os << "\tIsComplete: ";
  if (artsIpPathData.IsComplete())
    os << "true";
  else
    os << "false";
  os << endl;
  os << "\tNumHops: " << dec << (int)artsIpPathData.NumHops()
     << " (" << hex << (int)artsIpPathData.NumHops() << ")" << endl;

  vector<ArtsIpPathEntry>::const_iterator  pathEntry;
  for (pathEntry = artsIpPathData.Path().begin();
       pathEntry != artsIpPathData.Path().end();
       pathEntry++) {
    os << (*pathEntry);
  }
  
  return(os);
}

//----------------------------------------------------------------------------
//     void ArtsIpPathData::AddHop(ipv4addr_t ipAddr, uint8_t hopNum, 
//                                 const struct timeval & rtt) 
//............................................................................
//  
//----------------------------------------------------------------------------
void ArtsIpPathData::AddHop(ipv4addr_t ipAddr, uint8_t hopNum,
                            const struct timeval & rtt)
{
  ArtsIpPathEntry  pathEntry(ipAddr,hopNum);
  pathEntry.Rtt(rtt);
  this->_path.push_back(pathEntry);
  this->NumHops(this->Path().size());
}

uint32_t ArtsIpPathData::_numObjects = 0;
